from twomodalbert.Dataset import TwoModalDataset
from torch.utils.data import DataLoader
from sklearn.model_selection import train_test_split


class TwoModalDataPreparation:
    def __init__(
        self, config,
    ):
        self.max_seq_len = int(config["GENERAL"]["MAX_SEQ_LEN"])
        self.batch_size = int(config["GENERAL"]["BATCH_SIZE"])
        self.num_workers = int(config["GENERAL"]["NUM_WORKERS"])
        self.random_seed = int(config["GENERAL"]["RANDOM_SEED"])
        self.pretrained_model_name_or_path = config["GENERAL"][
            "PRETRAINED_MODEL_NAME_OR_PATH"
        ]

    def _create_data_loader(self, df, text_column, context_column, label_column):

        ds = TwoModalDataset(
            texts=df[text_column].to_numpy(),
            contexts=df[context_column].to_numpy(),
            labels=df[label_column].to_numpy(),
            max_seq_len=self.max_seq_len,
            pretrained_model_name_or_path=self.pretrained_model_name_or_path,
        )

        return DataLoader(ds, batch_size=self.batch_size, num_workers=self.num_workers)

    def prepare_data(
        self, df, text_column, context_column, label_column, train_size, val_size
    ):
        train, test = train_test_split(
            df,
            test_size=1 - train_size,
            random_state=self.random_seed,
            stratify=df[[label_column]],
        )
        test, val = train_test_split(
            test,
            test_size=val_size / (1 - train_size),
            random_state=self.random_seed,
            stratify=test[[label_column]],
        )

        train_data_loader = self._create_data_loader(
            train, text_column, context_column, label_column
        )
        val_data_loader = self._create_data_loader(
            val, text_column, context_column, label_column
        )
        test_data_loader = self._create_data_loader(
            test, text_column, context_column, label_column
        )
        return train_data_loader, train, val_data_loader, val, test_data_loader, test



import torch
from transformers import BertTokenizer
from torch.utils.data import Dataset


class TwoModalDataset(Dataset):
    def __init__(
        self,
        texts,
        contexts,
        labels,
        max_seq_len,
        pretrained_model_name_or_path="bert-base-uncased",
    ):
        self.texts = texts
        self.contexts = contexts
        self.labels = labels
        self.tokenizer = BertTokenizer.from_pretrained(pretrained_model_name_or_path)
        self.max_seq_len = max_seq_len

    def __len__(self):
        return len(self.texts)

    def tokenize(self, text):
        encoding = self.tokenizer.encode_plus(
            text=text,  # emails to encode.
            add_special_tokens=True,  # Add '[CLS]' and '[SEP]'
            max_length=self.max_seq_len,  # Pad & truncate all sentences.
            padding="max_length",
            return_attention_mask=True,  # Construct attn. masks.
            return_tensors="pt",  # Return pytorch tensors.
            truncation=True,
        )
        input_ids = encoding["input_ids"]
        attention_masks = encoding["attention_mask"]
        return input_ids, attention_masks

    def __getitem__(self, item):
        text = str(self.texts[item])
        context = str(self.contexts[item])
        label = self.labels[item]

        input_ids_text, attention_masks_text = self.tokenize(text)
        input_ids_context, attention_masks_context = self.tokenize(context)

        return {
            "text_text": text,
            "text_input_ids": input_ids_text.flatten(),
            "text_attention_mask": attention_masks_text.flatten(),
            "context_text": text,
            "context_input_ids": input_ids_context.flatten(),
            "context_attention_mask": attention_masks_context.flatten(),
            "labels": torch.as_tensor(label),
        }



from collections import defaultdict
from torch import nn
from transformers import AdamW
from twomodalbert.Model import TwoModalBERTModel
import numpy as np
import torch


class TwoModalBertTrainer:
    def __init__(self, device, config):
        self.device = device
        self.epochs = int(config["GENERAL"]["EPOCHS"])
        self.model_save_path = config["GENERAL"]["MODEL_SAVE_PATH"]
        self.pretrained_model_name_or_path = config["GENERAL"][
            "PRETRAINED_MODEL_NAME_OR_PATH"
        ]

    def _send_to_device(self, d):
        text_input_ids = d["text_input_ids"].to(self.device)
        text_attention_mask = d["text_attention_mask"].to(self.device)
        context_input_ids = d["context_input_ids"].to(self.device)
        context_attention_mask = d["context_attention_mask"].to(self.device)
        labels = d["labels"].to(self.device)
        return (
            text_input_ids,
            text_attention_mask,
            context_input_ids,
            context_attention_mask,
            labels,
        )

    def train_epoch(self, model, data_loader, loss_fn, optimizer, n_examples):
        model = model.train()

        losses = []
        correct_predictions = 0

        for d in data_loader:
            (
                text_input_ids,
                text_attention_mask,
                context_input_ids,
                context_attention_mask,
                labels,
            ) = self._send_to_device(d)

            outputs = model(
                text_input_ids,
                text_attention_mask,
                context_input_ids,
                context_attention_mask,
            )

            _, preds = torch.max(outputs, dim=1)

            loss = loss_fn(outputs, labels)

            correct_predictions += torch.sum(preds == labels)
            losses.append(loss.item())

            loss.backward()
            nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)
            optimizer.step()
            optimizer.zero_grad()

        return correct_predictions.double() / n_examples, np.mean(losses)

    def eval_model(self, model, data_loader, loss_fn, n_examples):
        model = model.eval()

        losses = []
        correct_predictions = 0

        with torch.no_grad():
            for d in data_loader:
                (
                    text_input_ids,
                    text_attention_mask,
                    context_input_ids,
                    context_attention_mask,
                    labels,
                ) = self._send_to_device(d)

                outputs = model(
                    text_input_ids,
                    text_attention_mask,
                    context_input_ids,
                    context_attention_mask,
                )

                _, preds = torch.max(outputs, dim=1)

                loss = loss_fn(outputs, labels)

                correct_predictions += torch.sum(preds == labels)
                losses.append(loss.item())

            return correct_predictions.double() / n_examples, np.mean(losses)

    def train_model(
        self,
        train_data_loader,
        train,
        val_data_loader,
        val,
        text_size,
        context_size,
        binary=False,
        text_p=0.3,
        context_p=0.3,
        output_p=0.3,
    ):
        history = defaultdict(list)
        print(f"LINE SIZE = {text_size}, CONTEXT SIZE = {context_size}")
        model = TwoModalBERTModel(
            text_size=text_size,
            context_size=context_size,
            binary=binary,
            text_p=text_p,
            context_p=context_p,
            output_p=output_p,
            pretrained_model_name_or_path=self.pretrained_model_name_or_path,
        )
        model = model.to(self.device)

        optimizer = AdamW(model.parameters(), lr=2e-5, correct_bias=False)

        loss_fn = nn.CrossEntropyLoss().to(self.device)

        best_accuracy = 0

        for epoch in range(self.epochs):

            print(f"Epoch {epoch + 1}/{self.epochs}")
            print("-" * 10)

            train_acc, train_loss = self.train_epoch(
                model, train_data_loader, loss_fn, optimizer, len(train)
            )

            print(f"Train loss {train_loss} accuracy {train_acc}")

            val_acc, val_loss = self.eval_model(
                model, val_data_loader, loss_fn, len(val)
            )

            print(f"Val loss {val_loss} accuracy {val_acc}")

            print()

            history["train_acc"].append(train_acc)
            history["train_loss"].append(train_loss)
            history["val_acc"].append(val_acc)
            history["val_loss"].append(val_loss)

            if val_acc > best_accuracy:
                torch.save(
                    model.state_dict(), self.model_save_path,
                )
                best_accuracy = val_acc

        return model, history



from configparser import ConfigParser
from transformers import BertTokenizer
import torch

config = ConfigParser()
config.read("config.ini")


def test_model(model, data_loader, device):
    model = model.eval()

    predictions = []
    real_values = []

    with torch.no_grad():
        for d in data_loader:
            text_input_ids = d["text_input_ids"].to(device)
            text_attention_mask = d["text_attention_mask"].to(device)
            context_input_ids = d["context_input_ids"].to(device)
            context_attention_mask = d["context_attention_mask"].to(device)
            labels = d["labels"].to(device)

            outputs = model(
                text_input_ids,
                text_attention_mask,
                context_input_ids,
                context_attention_mask,
            )

            _, preds = torch.max(outputs, dim=1)

            predictions.extend(preds)
            real_values.extend(labels)

    return predictions, real_values


def tokenize(text, device):
    tokenizer = BertTokenizer.from_pretrained(config["GENERAL"]["PRETRAINED_MODEL_NAME_OR_PATH"])
    encoding = tokenizer.encode_plus(
        text=text,  # emails to encode.
        add_special_tokens=True,  # Add '[CLS]' and '[SEP]'
        max_length=int(
            config["GENERAL"]["MAX_SEQ_LEN"]
        ),  # Pad & truncate all sentences.
        padding="max_length",
        return_attention_mask=True,  # Construct attn. masks.
        return_tensors="pt",  # Return pytorch tensors.
        truncation=True,
    )
    input_ids = encoding["input_ids"].to(device)
    attention_masks = encoding["attention_mask"].to(device)
    return input_ids, attention_masks


def predict_on_text(model, text, context, device):
    input_ids_text, attention_masks_text = tokenize(text, device)
    input_ids_context, attention_masks_context = tokenize(context, device)

    outputs = model(
        input_ids_text, attention_masks_text, input_ids_context, attention_masks_context
    )

    _, preds = torch.max(outputs, dim=1)

    return int(preds[0])



from torch import nn
from transformers import BertForSequenceClassification
import torch


class TwoModalBERTModel(nn.Module):
    def __init__(
        self,
        text_size,
        context_size,
        binary=False,
        text_p=0.3,
        context_p=0.3,
        output_p=0.3,
        pretrained_model_name_or_path="bert-base-uncased",
    ):
        super(TwoModalBERTModel, self).__init__()
        self.text_hs_size = text_size
        self.context_hs_size = context_size
        self.model = BertForSequenceClassification.from_pretrained(
            pretrained_model_name_or_path, output_hidden_states=True
        )
        # Reduce hidden state size of Bert output:
        self.linear_text = nn.Linear(self.model.config.hidden_size, self.text_hs_size)
        self.linear_context = nn.Linear(
            self.model.config.hidden_size, self.context_hs_size
        )
        self.dropout_text = nn.Dropout(p=text_p)
        self.dropout_context = nn.Dropout(p=context_p)
        # Add Linear layer that concatenates text and context embeddings
        self.linear_categorical_output = nn.Linear(
            self.text_hs_size + self.context_hs_size, 12
        )
        self.dropout_last = nn.Dropout(p=output_p)
        if binary:
            self.activation = nn.Sigmoid(dim=1)
        else:
            self.activation = nn.Softmax(dim=1)

    def create_mode(self, input_ids, attention_mask, layer_to_reduce_hs, dropout):
        outputs = self.model(input_ids=input_ids, attention_mask=attention_mask)
        last_hidden_states = outputs["hidden_states"][-1]
        cls_hidden_states = last_hidden_states[:, 0, :]
        last_hs_reduced = layer_to_reduce_hs(cls_hidden_states)
        last_hs_reduced = dropout(last_hs_reduced)
        return last_hs_reduced

    def forward(
        self,
        text_input_ids,
        text_attention_mask,
        context_input_ids,
        context_attention_mask,
    ):
        mode_text = self.create_mode(
            input_ids=text_input_ids,
            attention_mask=text_attention_mask,
            layer_to_reduce_hs=self.linear_text,
            dropout=self.dropout_text,
        )
        mode_context = self.create_mode(
            input_ids=context_input_ids,
            attention_mask=context_attention_mask,
            layer_to_reduce_hs=self.linear_context,
            dropout=self.dropout_context,
        )
        concatenated_modes = torch.cat([mode_text, mode_context], dim=1)
        categorical_output = self.linear_categorical_output(concatenated_modes)
        categorical_output = self.dropout_last(categorical_output)
        activation_output = self.activation(categorical_output)
        return activation_output



from sklearn import preprocessing
from sklearn.metrics import confusion_matrix
from src.DataPreparation import TwoModalDataPreparation
from src.Trainer import TwoModalBertTrainer
import os
import pandas as pd
import torch

from src.Model import TwoModalBERTModel
from configparser import ConfigParser
from src.Inference import predict_on_text, test_model

import seaborn as sns
import matplotlib.pyplot as plt

# load data
src_df_path = "/content/drive/MyDrive/TheOffice/The-Office-Lines-V4.csv"
src_df = pd.read_csv(src_df_path)
src_df["context"] = src_df["line"].shift(1)
df = src_df[["line", "context", "speaker"]]
SELECTED_SPEAKERS = [
    "Michael",
    "Jim",
    "Pam",
    "Dwight",
    "Jan",
    "Phyllis",
    "Stanley",
    "Oscar",
    "Angela",
    "Kevin",
    "Ryan",
    "Creed",
]
df = df[df["speaker"].isin(SELECTED_SPEAKERS)]
le = preprocessing.LabelEncoder()
df["label"] = le.fit_transform(df["speaker"])


# create data loaders
(
    train_data_loader,
    train,
    val_data_loader,
    val,
    test_data_loader,
    test,
) = DataPreparation.prepare_data(
    df,
    text_column="line",
    context_column="context",
    label_column="label",
    train_size=0.8,
    val_size=0.1,
)


# train model
model, history = Trainer.train_model(
    train_data_loader,
    train,
    val_data_loader,
    val,
    text_size=200,
    context_size=200,
    binary=False,
    text_p=0.3,
    context_p=0.3,
    output_p=0.3,
)


# load model
model = TwoModalBERTModel(
    text_size=200,
    context_size=200,
    binary=False,
    text_p=0.3,
    context_p=0.3,
    output_p=0.3,
)
model.load_state_dict(torch.load(config["GENERAL"]["MODEL_SAVE_PATH"]))


# evaluate on a test set
y_pred, y_test = test_model(model, test_data_loader)
y_pred, y_test = [e.cpu() for e in y_pred], [e.cpu() for e in y_test]


# print confusion matrix - helper function
def show_confusion_matrix(confusion_matrix):
    hmap = sns.heatmap(confusion_matrix, annot=True, fmt="d", cmap="Blues")
    hmap.yaxis.set_ticklabels(hmap.yaxis.get_ticklabels(), rotation=0, ha="right")
    hmap.xaxis.set_ticklabels(hmap.xaxis.get_ticklabels(), rotation=30, ha="right")
    plt.ylabel("True speaker")
    plt.xlabel("Predicted speaker")


cm = confusion_matrix(y_test, y_pred)
speakers_labels = (
    df[["speaker", "label"]].groupby(["speaker"]).agg("max").to_dict()["label"]
)
df_cm = pd.DataFrame(cm, index=speakers_labels, columns=speakers_labels)
show_confusion_matrix(df_cm)


# run on new pair of texts
line = "Dwight is my best friend."
context = "What do you think about Dwight?"

predict_on_text(model, line, context)
# output 12

speakers_mapping = {y: x for x, y in speakers_labels.items()}
speakers_mapping[12]
